declare('K_HTTP', {})
K_HTTP.user_agent = "TGFT Keys Client"

local wanted_headers = {
    location="Location",
    poweredby="X-Powered-By",
    contenttype="Content-type",
    server="Server",
}

local function unescape (s)
    s = string.gsub(s, "+", " ")
    s = string.gsub(s, "%%(%x%x)", function (h)
        return string.char(tonumber(h, 16))
    end)
    return s
end

local function escape (s)
    s = string.gsub(s, "([&=+%c])", function (c)
        return string.format("%%%02X", string.byte(c))
    end)
    s = string.gsub(s, " ", "+")
    return s
end

local function encode (t)
    local s = ""
    for k,v in pairs(t) do
        s = s .. "&" .. escape(k) .. "=" .. escape(v)
    end
    return string.sub(s, 2)     -- remove first `&'
end

local function decode (s)
    local cgi = {}
    for name, value in string.gfind(s, "([^&=]+)=([^&=]+)") do
        name = unescape(name)
        value = unescape(value)
        cgi[name] = value
    end
    return cgi
end

function K_HTTP.urlopen (url, method, callback, postdata)
    local body = ""
    local header = {}
    header.status = false
    local buffer = ""
    local to_body = false
    local active = false
    local appended = false
    local _, host, path, sock, type, length, rest
    local port = 80
    local gotit = false
    local nolen = false
    local postthis = ""
    if (method == nil) then method = "GET" end

    if not (string.find(url, "http://(.-)/(.*)$")) then url = url..'/' end
    _,_,host,path = string.find(url, "http://(.-)/(.*)$")
    if string.find(host, ':') then
        _, _, host, port = string.find(host, "(.*):(.*)$")
        port = tonumber(port)
    end


    local function callcallback(suc, hd, pg)
        active = false
        if (sock) and (sock.tcp) then
            sock.tcp:Disconnect()
        end
        sock = nil
        if not (callback == nil) then
            return callback(suc, hd, pg)
        end
    end

    if not (postdata == nil) then
        postthis = encode(postdata)
        if (method == "GET") then
            path = path.."?"..postthis
        end
    elseif method == "POST" then
        return callback("I need something to post", nil, nil)
    end

    local request = ""
    request = request..method.." /"..path.." HTTP/1.1\r\n"
    request = request.."Host: "..host..":"..tostring(port).."\r\n"
    request = request.."User-Agent: "..K_HTTP.user_agent.."\r\n"
    request = request.."Accept: text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5\r\n"
    request = request.."Accept-Language: en-us,en;q=0.5\r\n"
    request = request.."Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7\r\n"
    request = request.."Connection: close\r\n"

    if method == "POST" then
        request = request.."Content-Type: application/x-www-form-urlencoded\r\n"
        request = request.."Content-Length: "..string.len(postthis).."\r\n"
        request = request.."\r\n"..postthis
    else
        request = request.."\r\n"
    end

    local function ConnectionTimeOut()
        if active and sock then
            return callcallback("Connection timed out", nil, nil)
        end
    end

    local function ConnectionMade(con, suc)
        if not (suc == nil) then
            return callcallback(suc, nil, nil)
        else
            con:Send(request)
            local t = Timer()
            active = true
            t:SetTimeout(15000, ConnectionTimeOut)
        end
    end
    local function LineReceived(con, line)

        if line == '\r' then
      to_body = true
        end
        if not header.status and string.find(line, "^HTTP") then
            header.status = tonumber(string.sub(line, 10, 10+3))
        end

        if to_body then
            body = body..line..'\n'
            local curlen = string.len(body)
            if not (length == nil) then
                if curlen == length or curlen > length or length == 0 then
                    gotit = true
                    return callcallback(false, header, body)
                end
            else
                if not nolen then
                    print('-- don\'t know the length, waiting for the connection to get closed by the webserver')
                    nolen = true
                end
            end
        else
            local var, val
            _, _, var, val = string.find(line, "(.*): (.*)")
            if not (var == nil) and not (val == nil) then
                for wanted, real in pairs(wanted_headers) do
                    if var == real then
                        header[wanted] = val
                    end
                end
            end

            if string.find(line, "^Content(.*)Length") and length == nil then
                _, _, _, length = string.find(line, "Content(.*): (.*)$")
                length = tonumber(length)
            end
        end
    end
    local function ConnectionLost(con)
        --print('lost connection')
        if (header.status) then
            return callcallback(false, header, body)
        else
            return callcallback("Unknown error", nil, nil)
        end
    end
    sock = TCP.make_client(host, port, ConnectionMade, LineReceived, ConnectionLost)
end
